//------------------------------------------------------------
// This program was made by me, Emilio P.G. Ficara - Italy
// website: http://sites.google.com/site/eficara
// If you find something interesting in this site and use part
//  of it in your projects, please add a reference to my work
//  in your documentation, thank you !
//------------------------------------------------------------
#pragma language=extended       

#include <io2313.h>        
#include <ina90.h>

#define FW_VERSION  5.611_  // firmware release

/*********************************** 
*        Start of variables        * 
***********************************/

#define USER_RAM_LENGTH 17 // bytes of ram used for variables

unsigned char ledscan[4]; // led scan ram image
unsigned char ledbar; // active led bar
unsigned char presca; // prescaler for 1 second
unsigned char second; // seconds counter
unsigned char minutes; // minutes counter
unsigned char hours; // hours counter
unsigned char alarmhh; // alarm hour
unsigned char alarmmm; // alarm minute
unsigned char alarmflg; // alarm status ON./OFF
unsigned char kfase; // machine state for keyboard
unsigned char alarmreg; // flag for alarm setting state
unsigned char alarmnum; // counter for successive alarms
unsigned int kbounce; // key debounce counter

/*********************************** 
*        Start of constants        * 
***********************************/
flash unsigned char kmask[4] = { 0xfd, 0xfe, 0xf7, 0xfb };
flash unsigned char ledmap[6] = { 0xfe, 0xfd, 0xfb, 0xf7, 0xef, 0xdf };

/*********************************** 
*        Interrupts                * 
***********************************/

interrupt [TIMER0_OVF_vect] void TIMER0_OVF_interrupt(void) 
{  
    PORTD |= 0x0f; // all LED anodes OFF
    PORTB |= 0x3f; // all LED catodes OFF
    if((PIND & 0x70) != 0x70) // if any key pressed
        ++kbounce; // increment debounce counter
    else // no key pressed
        kbounce = 0; // reset debounce counter    
    PORTB &= ledscan[ledbar];
    PORTD &= kmask[ledbar];
    if(++ledbar >= 4)
        ledbar = 0;
    if(++presca >= 250) // if 1 second elapsed
        {
        presca = 0; // reset prescaler
        alarmflg |= 0x80; // flag for alarm led blink
        if(++second >= 60) // if 60 seconds elapsed
            {
            second = 0; // reset seconds
            if(++minutes >= 60) // if 60 minutes elapsed
                {
                minutes = 0; // reset minutes
                if(++hours >= 12) // if 12 hours elapsed
                    hours = 0; // reset hours
                }
            }
        }
    if(alarmflg & 0x40) // if ringing
        PORTB ^= 0x40; // toggle motor control (50% duty)
}

/*********************************** 
*        Main Program              * 
***********************************/

/* I/O ports map

    PD.7    *PD.6   *PD.5   *PD.4   PD.3    PD.2    PD.1   PD.0
     //     SW_HR   SW_MN   SW_AL   LDA3    LDA2    LDA1   LDA0

    PB.7    PB.6    PB.5    PB.4    PB.3    PB.2    PB.1    PB.0
    ALARM   MOTOR   LDK5    LDK4    LDK3    LDK2    LDK1    LDK0                                
*/                      

void InitAll(void)
{
unsigned char c;

    PORTD = 0x7f;   // 0111.1111
    DDRD = 0x0f;    // 0000.1111
    PORTB = 0xbf;   // 1011.1111
    DDRB = 0xff;    // 1111.1111

    TCCR0 = 0x03;   // T0 runs at clk/64 (256 counts overflow in 4 mS @ 4.096 MHz)
    TIMSK = 0x02;   // T0 overflow Int enable
    
    for(c = 0; c < USER_RAM_LENGTH; c++) // init all variables
        ledscan[c] = 0;

    alarmmm = 10; // default alarm minute (00:10)
    alarmflg = 1; // default alarm ON
    _SEI();
}

void ShowTime(void)
{
unsigned char tmn, thr;

    if(alarmreg) // if must show alarm time
        {
        thr = alarmhh; // get alarm hour in the range 0..11
        tmn = alarmmm / 5; // get alarm minutes in the range 0..11
        }
    else // if show normal time
        {    
        thr = hours; // get hours in the range 0..11
        tmn = minutes / 5; // get minutes in the range 0..11
        }
        
    if(thr < 6) // if hour is 0..5
        {
        ledscan[2] = ledmap[thr]; // update led map
        ledscan[3] = 0xff; // update led map
        }
    else // if hour is 6..11
        {
        ledscan[2] = 0xff; // update led map
        ledscan[3] = ledmap[thr-6]; // update led map
        }
            
    if(tmn < 6) // if minutes are 00..25
        {
        ledscan[0] = ledmap[tmn]; // update led map
        ledscan[1] = 0xff; // update led map
        }
    else // if minutes are 30..55
        {
        ledscan[0] = 0xff; // update led map
        ledscan[1] = ledmap[tmn-6]; // update led map
        }
}

void CheckAlarm(void)
{
    if(alarmflg & 1) // if alarm set
        {
        if(alarmflg & 0x80) // if 1 second elapsed
            {
            alarmflg &= ~0x80; // turn off flag
            PORTB &= ~0x80; // turn ON led
            }
        else
            {
            if(presca > 25 && (PORTB & 0x80) == 0) // if 100 ms with led on
                PORTB |= 0x80; // turn OFF led
            }
            
        if(hours == alarmhh && minutes == alarmmm) // if alarm time
            {
            PORTB &= ~0x80; // turn ON alarm led
            _CLI();
            alarmflg |= 0x40; // set flag for ringing
            _SEI();
            for(;;)
                {
                if((PIND & 0x10) == 0) // if SW_AL pressed
                    break; // exit for (stop ringing)
                if(minutes != alarmmm) // if 1 minute elapsed
                    break; // exit for (stop ringing)
                ShowTime(); // update time
                }
            PORTB |= 0x80; // turn OFF alarm led
            _CLI();
            alarmflg &= ~0x41; // turn OFF ringer and alarm flag
            PORTB &= ~0x40; // stop ringing
            _SEI();
            if(--alarmnum > 0) // if not last alarm
                {
                _CLI();
                alarmflg |= 1; // set again the alarm
                _SEI();
                alarmmm += 5; // set next alarm in 5 minutes
                if(alarmmm > 55)
                    {
                    alarmmm = 0;
                    if(++alarmhh > 11)
                        alarmhh = 0;
                    }
                }
            }
        }
}

void CheckKeys(void)
{
    switch(kfase) // parse machine state for keyboard
        {
        case 0: // waiting for key pressed
            if(kbounce > 25) // if key pressed at least 100 mS
                {
                if((PIND & 0x10) == 0) // if 'alarm' pressed
                    kfase = 3; // machine state = view alarm time / alarm set
                else if((PIND & 0x20) == 0) // if 'min' pressed
                    kfase = 1; // machine state = minutes advance
                else if((PIND & 0x40) == 0) // if 'hour' pressed
                    kfase = 2; // machine state = hours advance
                }
            break;
        case 1: // minutes advance
            if(kbounce <= 5) // if key released
                kfase = 0; // machine state = waiting for key pressed
            if(kbounce >= 500) // if 2 sec pressed (first time, otherwise 700 mS pressed)
                {
                _CLI(); // disable interrupts                
                kbounce = 325; // repeat in 700 mS
                minutes = (minutes / 5) * 5 + 5;
                if(minutes > 55)
                    minutes = 0;
                _SEI(); // enable interrupts
                }
            break;    
        case 2: // hours advance
            if(kbounce <= 5) // if key released
                kfase = 0; // machine state = waiting for key pressed
            if(kbounce >= 500) // if 2 sec pressed (first time, otherwise 700 mS pressed)
                {
                _CLI(); // disable interrupts                
                kbounce = 325; // repeat in 700 mS
                ++hours;
                if(hours > 11)
                    hours = 0;
                _SEI(); // enable interrupts
                }
            break;    
        case 3: // view alarm time
            alarmreg = 1; // set alarm time visualization
            if(kbounce <= 5) // if key released
                {
                kfase = 0; // machine state = waiting for key pressed
                alarmreg = 0; // restore normal time visualization
                }
            if(kbounce >= 500) // if 2 sec pressed (first time, otherwise 700 mS pressed)
                {
                kfase = 4; // machine state = alarm set
                alarmnum = 3; // 3 times in alarm (5 minutes interval)
                _CLI(); // disable interrupts                
                alarmflg |= 1; // turn ON alarm flag
                 _SEI(); // enable interrupts
                }
            break;
        case 4: // alarm set
            if(kbounce <= 5 || (PIND & 0x10)) // if key released or 'alarm' nor pressed
                {
                kfase = 0; // machine state = waiting for key pressed
                alarmreg = 0; // restore time visualization
                }
            PORTB &= ~0x80; // force alarm led ON
            if(kbounce > 500) // if 2 sec pressed (first time, otherwise 700 mS pressed)
                {
                _CLI(); // disable interrupts                
                kbounce = 325; // repeat in 700 mS
                 _SEI(); // enable interrupts
                if((PIND & 0x20) == 0) // if 'min' pressed
                    {
                    ++alarmhh;
                    if(alarmhh > 11)
                        alarmhh = 0;
                    }
                if((PIND & 0x40) == 0) // if 'hour' pressed
                    {
                    alarmmm = (alarmmm / 5) * 5 + 5;
                    if(alarmmm > 55)
                        alarmmm = 0;
                    }
                }
            break;
        }
}

void C_task main(void)
{ 
    InitAll(); // init registers, clear ram, enable interrupts

    for(;;)  
        {    
        ShowTime();
        CheckAlarm();
        CheckKeys();
        } 
}
